/* Header file for the Lexical Analyser Generator (dlg.c)
 * for ee595e
 *
 * Will Cohen
 * 9/27/89
 */

#include	"sets.h"

#define TRUE	1
#define FALSE	0

#define IN	infile
#define OUT	outfile

/* valid characters (classes) for input */
#define FIRST_CHAR	0
#define LAST_CHAR	127
#define NUM_CLASS	(LAST_CHAR-FIRST_CHAR+1+1) /* 1 for LEX_EOF */

#define MAX_STACK	64	/* depth of stack for building nfa's */

#define nfa_node struct _nfa_node
nfa_node {
	int		nfa_set;
	int		accept;	/* what case to use */
	int		trans[2];
	set		label;	/* one arc always labelled with epsilon */
	};

#define dfa_node struct _dfa_node
dfa_node {
	int		done;
	set		nfa_states;
	int		dfa_set;
	short		trans[NUM_CLASS];
	};

#define path	struct _path
path {
	int	l;
	int	r;
	};

/* stuff to manage nfa and dfa graphs */

extern int	nfa_pool;	/* last nfa node made */
extern int	dfa_pool;	/* last dfa node made */
extern nfa_node	*nfa_array;	/* start of allocated space for nfa */
extern dfa_node	*dfa_array;	/* start of allocated space for dfa */
extern int	*dfa_hash_table;

/* definitions to make access nodes in dfa and nfa transparent to implentation
 * (not fully implemented )
 */
#define NIL_INDEX	-1	/* impossible index for array */

#define PUSH_PATH(x)	(stack[sp++] = (x))
#define POP_PATH()	(stack[--sp])
#define TOS		(stack[sp-1])
#define NOS		(stack[sp-2])

#define nfa(x)		(nfa_array[(x)])

/* this macro was made becase the sequence
	e = set_int(s);
	s = set_dif(s,set_of(e));
  was heavily used in closure() and nfa_to_dfa()
  If e is a element in the set, starts searching for elements in word or
  higher in set to save some time.
*/
#define set_remove_int(b,e,s) {\
	/* Fast pick any element of the set b starting at the word pointed\
	   to by s. s needs to be a real variable for this to work */\
	register unsigned *p;\
	register unsigned *endp = &(b.setword[SETWORDS]);\
	register int i;\
	register int nothing = TRUE;\
\
	do {\
		if (*s) {\
			/* Found a non-empty word of the set */\
			register unsigned t = *s;\
			i = ((s - &(b.setword[0])) * WORDSIZE);\
			p = &(bitmask[0]);\
			while ((nothing = !(*p & t))) {\
				++i; ++p;\
				}\
			}\
		} while (nothing && (++s < endp));\
		if (!nothing){\
		    b.setword[DIVWORD(i)] &= (~bitmask[MODWORD(i)]);\
		    e = i;\
		    }\
		else\
		    e = nil;\
}

/* found that set_or(s,set_or(e)); common (doesn't work for nil)*/
#define set_or_int(s,e) (s.setword[ DIVWORD(e)] |= bitmask[MODWORD(e)])

			
/***************************** data ***********************************/
extern int	token;
extern int	line_no;
extern nfa_node	clear_nfa_node;	/* used to initialize new nfas */
extern dfa_node clear_dfa_node;	/* used to initialize new dfas */
extern path	stack[];	/* pass back node list for building nfa */
extern int	sp;		/* stack pointer for above */
extern set	used_chars;	/* used to label trans. arcs */
extern set	used_classes;	/* classes or chars used to label trans. arcs */
extern set	valid_chars;	/* mask off unused portion of set */
extern int	nfa_set;
extern int	num_of_nfa;
extern int	num_of_dfa;
extern int	hash_entries;	/* number of possible entries in hash table */
extern int	compress;	/* flag for character class compression */
extern int	action_no;	/* number of actions produced */
extern unsigned	bitmask[];	/* array used to speed up set operations */
extern int	class_no;	/* number of classes actually used */
extern set	class[];	/* stores partitions of classes */

/*************************** functions ********************************/
extern	void	init_things();
extern	void	error();
extern	void	warning();
extern	void	fatal();
extern	set	reach();
extern	set	closure();

/************************** file stuff *******************************/
extern	FILE	*infile;
extern	FILE	*outfile;
extern	char	*in_file_name;	/* remembers name for error reporting */
